from pathlib import Path
from django.test import TestCase
from django.core import mail

from ichec_django_core.models import Member

from ichec_django_core.test import (
    setup_default_users_and_groups,
)

from marinerg_test_access.test import (
    setup_access_call,
    setup_application,
)


class ApplicationSummaryTestCase(TestCase):

    def setUp(self):
        setup_default_users_and_groups()
        call = setup_access_call()

        test_data_dir = Path(__file__).parent / "data"
        self.application = setup_application(
            call, "test_application", test_data_dir / "test_application.pdf"
        )

    def test_summary_generation(self):
        self.application.status = "SUBMITTED"
        self.application.save()



from django.utils import timezone

from ichec_django_core.models import Member

from ichec_django_core.test import (
    AuthAPITestCase,
    setup_default_users_and_groups,
)

from marinerg_test_access.test import setup_access_call


class AccessCallViewTests(AuthAPITestCase):

    template = {
        "title": "My Access Call",
        "description": "Description of my access call",
        "coordinator": "",
        "board_chair": "",
        "board_members": [],
        "closing_date": str(timezone.now()),
        "form": {
            "groups": [
                {
                    "fields": [
                        {
                            "label": "Requires Support",
                            "key": "requires_support",
                            "description": "Do you require facility support?",
                            "field_type": "BOOLEAN",
                        }
                    ]
                }
            ]
        },
    }

    update_template = {"description": "Modified description of my access call"}

    def setUp(self):
        self.url = "/api/access_calls/"

        setup_default_users_and_groups()

        setup_access_call()

    def get_user(self, name: str) -> Member:
        return Member.objects.get(username=name)

    def get_user_url(self, name: str):
        return f"http://testserver/api/members/{self.get_user(name).id}/"

    def get_template(self, username: str = "admin_user") -> dict:
        template = self.template
        template["coordinator"] = self.get_user_url(username)
        template["board_chair"] = self.get_user_url(username)
        return template

    def check_response(self, returned, expected, skip_keys: tuple = ()):
        for key, expected_value in expected.items():

            if key in skip_keys:
                continue

            returned_value = returned[key]
            if isinstance(expected_value, dict):
                self.compare_dicts(returned_value, expected_value, skip_keys)
            elif isinstance(expected_value, list):
                self.assertEqual(len(expected_value), len(returned_value))
            else:
                self.assertEqual(returned_value, expected_value)

    def test_list_not_authenticated(self):
        self.assert_401(self.do_list())

    def test_detail_not_authenticated(self):
        self.assert_401(self.detail(1))

    def test_list_regular_user(self):
        self.assert_200(self.authenticated_list("regular_user"))

    def test_detail_regular_user(self):
        self.assert_200(self.authenticated_detail("regular_user", 1))

    def test_create_not_authenticated(self):
        self.assert_401(self.create(self.get_template()))

    def test_create_regular_user(self):
        self.assert_403(self.authenticated_create("regular_user", self.get_template()))

    def test_create_consortium_admin(self):
        response = self.assert_201(
            self.authenticated_create("admin_user", self.get_template())
        )
        # self.check_response(response, self.get_template(), ("closing_date"))

    def test_update_consortium_admin(self):
        created = self.assert_201(
            self.authenticated_create("admin_user", self.get_template())
        )

        for key, value in self.update_template.items():
            created[key] = value

        created["form"]["groups"][0]["fields"].append(
            {
                "label": "Requires More Support",
                "key": "requires_more_support",
                "description": "Do you require more facility support?",
                "field_type": "BOOLEAN",
            }
        )

        updated = self.assert_200(
            self.authenticated_update("admin_user", created["id"], created)
        )
        self.assertTrue(
            updated["description"] == "Modified description of my access call"
        )
        self.assertTrue(len(updated["form"]["groups"][0]["fields"]) == 2)



from pathlib import Path

from django.utils import timezone

from ichec_django_core.models import (
    Member,
    Form,
    FormGroup,
    FormField,
    PopulatedForm,
    FormFieldValue,
)

from ichec_django_core.test import add_group_permissions

from marinerg_test_access.models import AccessCall, AccessApplication


def setup_access_call():

    coordinator = Member.objects.create(
        username="access_call_coordinator",
        first_name="Access Call",
        last_name="Coordinator",
        email="access_call_coordinator@example.com",
    )
    access_call_board_member = Member.objects.create(
        username="access_call_board_member",
        first_name="Access Call",
        last_name="Board Member",
        email="access_call_board_member@example.com",
    )

    form = Form.objects.create()

    default_group = FormGroup.objects.create(form=form)

    FormField.objects.create(
        group=default_group,
        label="Sample Bool Field?",
        key="bool_field",
        required=True,
        field_type="BOOLEAN",
    )

    FormField.objects.create(
        group=default_group,
        label="Application Form",
        key="application_form",
        required=True,
        field_type="FILE",
    )

    access_call = AccessCall.objects.create(
        title="Test access call",
        description="Description of access call",
        status="OPEN",
        closing_date=timezone.now(),
        coordinator=coordinator,
        board_chair=access_call_board_member,
        form=form,
    )

    access_call.board_members.set([access_call_board_member])

    add_group_permissions("admins", AccessCall, ["change_accesscall", "add_accesscall"])
    return access_call


def setup_application(call, applicant="test_applicant", asset: Path | None = None):

    form = PopulatedForm.objects.create()

    FormFieldValue.objects.create(
        value="TRUE", field=call.get_field_by_key("bool_field"), form=form
    )

    if asset:
        file_field = FormFieldValue.objects.create(
            field=call.get_field_by_key("application_form"), form=form
        )
        file_field.asset.original_filename = asset.name
        file_field.asset.name = str(asset)
        file_field.save()

    applicant = Member.objects.create(
        username=applicant,
        first_name="Test",
        last_name="Applicant",
        email=applicant + "@example.com",
    )
    return AccessApplication.objects.create(applicant=applicant, call=call, form=form)



import uuid
import os
from pathlib import Path

from django.http import Http404
from django.core.files.storage import FileSystemStorage
from django.conf import settings

from rest_framework import permissions
from rest_framework.decorators import action
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.parsers import FileUploadParser

from ichec_django_core.models import Member, FormFieldValue
from ichec_django_core.views import (
    ObjectFileDownloadView,
    SearchableModelViewSet,
)

from marinerg_test_access.models import AccessApplication
from marinerg_test_access.serializers import (
    AccessApplicationCreateSerializer,
    AccessApplicationDetailSerializer,
)


def is_consortium_admin(user):
    for group in user.groups.all():
        if group.name == "consortium_admins":
            return True
    return False


def is_on_access_board(application, user):
    if user.id == application.call.board_chair.id:
        return True

    for member in application.call.board_members.all():
        if member.id == user.id:
            return True
    return False


def is_call_coordinator(application, user):
    return application.call.coordinator.id == user.id


class ApplicationPermissions(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):

        if request.method in ["OPTIONS", "HEAD", "POST"]:
            return True

        if obj.applicant.id == request.user.id:
            return True
        if is_consortium_admin(request.user):
            return True

        if request.method in ["GET", "PUT", "PATCH"]:
            if is_on_access_board(obj, request.user):
                return True
            if is_call_coordinator(obj, request.user):
                return True
        return False


class AccessApplicationViewSet(SearchableModelViewSet):
    queryset = AccessApplication.objects.all()
    serializer_class = AccessApplicationDetailSerializer
    permission_classes = [
        permissions.IsAuthenticated,
        ApplicationPermissions,
    ]

    ordering: tuple[str, ...] = ("created_at",)
    search_fields = ["applicant__usename"]

    serializers = {
        "retrieve": AccessApplicationDetailSerializer,
        "list": AccessApplicationDetailSerializer,
        "create": AccessApplicationCreateSerializer,
        "update": AccessApplicationDetailSerializer,
        "partial_update": AccessApplicationDetailSerializer,
    }

    filterset_fields = (
        "applicant__id",
        "call__id",
        "facilities__id",
        "chosen_facility__id",
        "status",
    )

    @action(detail=True, url_path="summary", url_name="summary")
    def summary(self, request, pk=None):
        return self.download(request, "summary")

    def perform_create(self, serializer):
        serializer.save(applicant=Member.objects.get(id=self.request.user.id))


class ApplicationFieldDownloadView(ObjectFileDownloadView):

    model = AccessApplication
    permission_classes = [
        permissions.IsAuthenticated,
        ApplicationPermissions,
    ]

    def get_file(self):
        if not self.value.asset:
            raise Http404
        return self.value.asset

    def get_basename(self):
        ext = self.value.asset.name.split(".")[-1]
        return f"application_{self.value.field.key}.{ext}"

    def get(self, request, *args, **kwargs):
        self.authenticate(request)
        self.has_permission(request)

        self.object = self.get_object()
        self.value = FormFieldValue.objects.get(id=kwargs["value_id"])
        return super(ObjectFileDownloadView, self).get(request, *args, **kwargs)


class ApplicationFieldUploadView(GenericAPIView):

    model = AccessApplication
    queryset = AccessApplication.objects.all()
    parser_classes = [FileUploadParser]
    permission_classes = [
        permissions.IsAuthenticated,
        ApplicationPermissions,
    ]
    extensions = (".odt", ".doc", ".docx", ".pdf")

    def put(self, request, *args, **kwargs):
        value = FormFieldValue.objects.get(id=kwargs["value_id"])

        media_root = Path(settings.MEDIA_ROOT)
        instance = self.get_object()
        storage_dir = Path(f"{self.model._meta.model_name}/{instance.id}/field_values")
        os.makedirs(media_root / storage_dir, exist_ok=True)

        filename = request.data["file"].name

        _, ext = os.path.splitext(filename)
        if ext not in self.extensions:
            return Response(
                {"reason": "File extension not supported"},
                status=400,
            )

        file_path = f"{uuid.uuid4()}{ext}"
        written_path = FileSystemStorage(location=media_root / storage_dir).save(
            file_path, request.data["file"]
        )

        value.asset.original_file_name = filename
        value.asset.name = str(storage_dir / written_path)
        value.save()

        return Response(status=204)



from ichec_django_core.test import (
    AuthAPITestCase,
    setup_default_users_and_groups,
)

from marinerg_test_access.test import (
    setup_access_call,
    setup_application,
)


class AccessApplicationViewTests(AuthAPITestCase):

    template = {
        "call": "http://testserver/api/access_calls/1/",
        "form": {"values": []},
        "facilities": [],
        "dates_flexible": True,
    }

    def setUp(self):
        self.url = "/api/access_applications/"
        setup_default_users_and_groups()
        call = setup_access_call()
        setup_application(call, "test_applicant")

    def test_list_not_authenticated(self):
        self.assert_401(self.do_list())

    def test_detail_not_authenticated(self):
        self.assert_401(self.detail(1))

    def test_list_authenticated(self):
        self.assert_200(self.authenticated_list("regular_user"))

    def test_detail_regular_user(self):
        self.assert_403(self.authenticated_detail("regular_user", 1))

    def test_detail_test_applicant(self):
        self.assert_200(self.authenticated_detail("test_applicant", 1))

    def xtest_detail_consortium_admin(self):
        self.assert_200(self.authenticated_detail("admin_user", 1))

    def test_create_not_authenticated(self):
        self.assert_401(self.create(self.template))

    def test_create_regular_user(self):
        self.assert_201(self.authenticated_create("regular_user", self.template))

    def test_update_regular_user(self):
        response = self.authenticated_create("regular_user", self.template)
        self.assert_201(response)

        created = response.json()

        created["dates_flexible"] = False
        updated = self.authenticated_update("regular_user", created["id"], created)
        self.assert_200(updated)

    def test_submit(self):
        response = self.authenticated_create("regular_user", self.template)
        self.assert_201(response)

        created = response.json()

        created["status"] = "SUBMITTED"
        updated = self.authenticated_update("regular_user", created["id"], created)
        self.assert_200(updated)



import uuid
import os
from pathlib import Path

from django.conf import settings
from django.http import Http404
from django.core.files.storage import FileSystemStorage

from rest_framework import permissions, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.generics import GenericAPIView
from rest_framework.parsers import FileUploadParser

from ichec_django_core.models import FormField
from ichec_django_core.views import (
    SearchableModelViewSet,
    ObjectFileDownloadView,
    OwnerFullOrDjangoModelPermissions,
)

from marinerg_test_access.models import AccessCall, AccessCallFacilityReview
from marinerg_test_access.serializers import (
    AccessCallListSerializer,
    AccessCallDetailSerializer,
    AccessCallCreateSerializer,
    AccessCallFacilityReviewSerializer,
)


class CoordinatorEditOrDjangoModelPermissions(OwnerFullOrDjangoModelPermissions):
    owner_field = "coordinator"


class AccessCallViewSet(SearchableModelViewSet):
    queryset = AccessCall.objects.all()
    serializer_class = AccessCallListSerializer

    serializers = {
        "retrieve": AccessCallDetailSerializer,
        "list": AccessCallListSerializer,
        "create": AccessCallCreateSerializer,
        "update": AccessCallDetailSerializer,
        "partial_update": AccessCallDetailSerializer,
    }

    permission_classes = [
        permissions.IsAuthenticated,
        CoordinatorEditOrDjangoModelPermissions,
    ]

    ordering_fields = ("status",)
    ordering: tuple[str, ...] = ("-closing_date",)
    search_fields = ["title"]

    @action(
        detail=True, url_path="applications_summary", url_name="applications_summary"
    )
    def applications_summary(self, request, pk=None):
        return self.download(request, "applications_summary")


class AccessCallFieldDownloadView(ObjectFileDownloadView):

    model = AccessCall
    permission_classes = [
        permissions.IsAuthenticated,
        CoordinatorEditOrDjangoModelPermissions,
    ]

    def get_file(self):
        if not self.value.template:
            raise Http404
        return self.value.template

    def get_basename(self):
        ext = self.value.template.name.split(".")[-1]
        return f"access_call_{self.value.key}.{ext}"

    def get(self, request, *args, **kwargs):
        self.authenticate(request)
        self.has_permission(request)

        self.object = self.get_object()
        self.value = FormField.objects.get(id=kwargs["value_id"])
        return super(ObjectFileDownloadView, self).get(request, *args, **kwargs)


class AccessCallFieldUploadView(GenericAPIView):

    model = AccessCall
    queryset = AccessCall.objects.all()
    parser_classes = [FileUploadParser]
    permission_classes = [
        permissions.IsAuthenticated,
        CoordinatorEditOrDjangoModelPermissions,
    ]
    extensions = (".odt", ".doc", ".docx", ".pdf")

    def put(self, request, *args, **kwargs):
        value = FormField.objects.get(id=kwargs["value_id"])

        media_root = Path(settings.MEDIA_ROOT)
        instance = self.get_object()
        storage_dir = Path(
            f"{self.model._meta.model_name}/{instance.id}/field_templates"
        )
        os.makedirs(media_root / storage_dir, exist_ok=True)

        filename = request.data["file"].name

        _, ext = os.path.splitext(filename)
        if ext not in self.extensions:
            return Response(
                {"reason": "File extension not supported"},
                status=400,
            )

        file_path = f"{uuid.uuid4()}{ext}"
        written_path = FileSystemStorage(location=media_root / storage_dir).save(
            file_path, request.data["file"]
        )

        value.template.original_file_name = filename
        value.template.name = str(storage_dir / written_path)
        value.save()

        return Response(status=204)


class AccessCallFacilityReviewViewSet(viewsets.ModelViewSet):
    queryset = AccessCallFacilityReview.objects.all().order_by("id")
    serializer_class = AccessCallFacilityReviewSerializer
    permission_classes = [
        permissions.IsAuthenticated,
        permissions.DjangoModelPermissions,
    ]



from django.db import models

import functools

from ichec_django_core.models import (
    Member,
    Form,
    TimesStampMixin,
    DefaultPublicMixin,
    SoftDeleteMixin,
    content_file_name,
)

from marinerg_facility.models import Facility


class AccessCall(TimesStampMixin, DefaultPublicMixin, SoftDeleteMixin):

    class Meta:
        verbose_name = "Access Call"
        verbose_name_plural = "Access Calls"

    class StatusChoice(models.TextChoices):
        OPEN = "OPEN", "Open"
        CLOSED = "CLOSED", "Closed"
        DRAFT = "DRAFT", "Draft"

    title = models.CharField(max_length=200)
    description = models.TextField()
    status = models.CharField(
        max_length=10, choices=StatusChoice.choices, default=StatusChoice.DRAFT
    )
    closing_date = models.DateTimeField()
    coordinator = models.ForeignKey(Member, on_delete=models.CASCADE)

    board_chair = models.ForeignKey(
        Member, on_delete=models.CASCADE, related_name="chaired_access_boards"
    )
    board_members = models.ManyToManyField(
        Member,
        verbose_name="list of review board members",
        related_name="access_boards",
        blank=True,
    )
    selectable_facilities = models.ManyToManyField(
        Facility,
        verbose_name="Selectable Facilities",
        related_name="selectable_access_calls",
        blank=True,
    )

    applications_summary = models.FileField(
        null=True,
        upload_to=functools.partial(
            content_file_name, "access_call_applications_summary"
        ),
    )

    form = models.OneToOneField(Form, on_delete=models.CASCADE)

    def __str__(self):
        return self.title

    def save(self, force_insert=False, force_update=False, *args, **kwargs):

        if str(self.status).lower() == "closed":
            applications = self.applications.all()
            for application in applications:
                application.status = "BOARD"
                application.save()

        super().save(force_insert, force_update, *args, **kwargs)

    def get_field_by_key(self, key: str):
        for group in self.form.groups.all():
            for field in group.fields.all():
                if field.key == key:
                    return field
        raise RuntimeError(f"Requested field '{key}' not found")


class AccessCallFacilityReview(TimesStampMixin):

    class Meta:
        verbose_name = "Access Call Facility Review"
        verbose_name_plural = "Access Call Facility Reviews"

    class Decision(models.TextChoices):
        ACCEPT = "ACCEPT", "Accept"
        REJECT = "REJECT", "Reject"

    decision = models.CharField(max_length=10, choices=Decision.choices)
    comments = models.TextField(blank=True)
    call = models.ForeignKey(AccessCall, on_delete=models.CASCADE)
    facility = models.ForeignKey(Facility, on_delete=models.CASCADE)
